#!/usr/bin/perl


# use the CGI.pm module to access CGI functionality

	use CGI;
	$cgi = new CGI;


# write a generic header to prepare for text output

	$header = $cgi->header;
	print $header;
	

# get variables passed in from do_choose()

	$stored_page = $cgi->param('page');
	$stored_template = $cgi->param('template');
	$action = $cgi->param('action');


# define some useful subroutines

	sub get_file {
		my $file = $_[0]; my $method = $_[1]; my $check_last_modified = $_[2]; my $text = ""; # method can be "http", "filesystem", or "auto" (auto looks at $pre_render_files list)
		if (($file =~ /\/(\.\w.*)/)) { # don't read files whose names start with "."
			$text = show_error("Not authorized to read the hidden file named \"$1\"");
		} elsif (($file =~ /[^\w-\.\/\?\&\=\%]/)) { # don't read files whose names contain characters that could be used to read remote files, or to hack into the open command (Perl only)
			$text = show_error("File names can't contain the character $&");
		} else {
			
			# first, figure out if we're going to read over HTTP or through the filesystem...
			if (($method eq "")||($method eq "auto")) {
				if (@pre_render_files) {
					for ($n=0; $n < @pre_render_files; $n++) {
						if ($file =~ /@pre_render_files[$n]/) { $method = "http"; }
					}
				}
			}
			
			if ($method eq "http") {
				eval 'use LWP::Simple'; 
				if ($@) {
					$output .= show_error("Can\'t access Perl module LWP::Simple for pre-rendering files<br>Reading files directly instead<br>Hint: Set pre-render files to 0 in preferences.ini to skip this routine");
					$method = "filesystem";
				}
			}
			
			# ...then get the file
			if ($method eq "http") {
			
				# 1) prepare the URL for this type of connection
				$file_URL = absolute_path($file, "http");

				# 2) read the file
				$myfile = get($file_URL);
				$file_exists = 1; # to do: how can we test for this?
				if (! $file_exists) {
					$text = show_error("Can't open the file named \"" + $file + "\"" + $check_URL);
					$file_missing = 1;
				} else {
					$text = $myfile;
				}
				
				#$output .= "<p>reading $file over http<p>";  # debug
				
			} else {
			
				# 1) prepare the URL for this type of connection
				$file_URL = absolute_path($file, "filesystem");
				
				# 2) open a connection to the file
				open (myfile, $file_URL) or
				$text = show_error("Can't open the file named \"$file\"<br>$! $check_URL");
				if ($text =~ /Can't open/) {
					$file_missing = 1;
				} else {
				
					# 3) read the file
					while (<myfile>) { $text .= "$_"; }
					$file_missing = 0;
					
					# 4) for local files, catch the file's modification date (for $read_page_definitions_from and $template, plus content files when the template contains a env_last_modified tag)
					if ($check_last_modified) {
						($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = stat($file_URL);
						if ($mtime > $latest_mod_date) {
							$latest_mod_date = $mtime;
							($second, $minute, $hour, $day, $month, $year, $weekday, $yearday, $dst_flag) = localtime($latest_mod_date);
							$year += 1900;
							$month = sprintf("%02d",$month+1);
							$day = sprintf("%02d",$day);
							if ($hour < 12) {
								$ampm = "am";
							} else {
								$ampm = "pm";
								if ($hour > 12) {
									$hour = $hour - 12;
								}
							}
							#$hour = sprintf("%02d",$hour);
							$minute = sprintf("%02d",$minute);
							$latest_mod_date_formatted = "$year-$month-$day - $hour:$minute $ampm";
						}
					}

					# 5) close the file connection
					close (myfile);
					
				}
				
				#$output .= "<p>reading $script_location_copy + $file_copy through filesystem<p>";

			}
			
			# finally, conform all line breaks to avoid cross-platform issues
			# we'd rather conform to Unix line breaks (\n), but that doesn't work in ASP, which doesn't support the s flag in regular expressions
			$text =~ s/\r\n/\r/g; # change DOS to Mac
			$text =~ s/\n/\r/g; # change Unix to Mac
	
		}
		return $text;
	}
	
	sub get_page_names { # return a list of all the page names in a page definitions file
		my $source = $_[0];
		@page_names = ();
		while ($source =~ /([-\w]+)\:\s+[\w]+/) {
			$this_page_name = $1;
			push(@page_names, $this_page_name);
			$source =~ s/$this_page_name\://;
		}
		return @page_names;
	}

	sub get_setting {
		my $source_text = $_[0]; my $setting = $_[1]; my $allow_override = $_[2]; my $allow_partial = $_[3]; my $value = "";
		
		# allow most preferences to be overridden through the argument string
		$setting_no_spaces = $setting;
		$setting_no_spaces =~ s/[ -]/_/g;
		$settings_used .= "$setting_no_spaces,";

		$override = $cgi->param($setting_no_spaces);
		if (($allow_override)&&($override ne "")) {
			$value = $override;
			# don't allow users to pass code through the URL arguments
			$value =~ s/</&lt;/i;
			$value =~ s/>/&gt;/i;
		} else {
			$source_text = "\r".$source_text;  # this makes it easy to make sure settings start on a new line
			if ($allow_partial) {  # this lets us use more verbose labels in the ini file
				if ($source_text =~ /[^\w]$setting[^:\r]*: ([^\r]*)\r*/i) {
					$value = $1;
				}
			} else {  # this ensures an exact match in the pages file
				if ($source_text =~ /[^\w]$setting: ([^\r]*)\r*/i) {
					$value = $1;
				}
			}
		}
		if ($value eq "0") { $value = 0; } # this ensures that a 0 value will evaluate false in every language
		return $value;
	}
	
	sub get_directory_listing { # return a list of all the items in the specified directory
		my $directory = $_[0];
		opendir(dir, absolute_path($directory, "filesystem")) or
		show_error("Can't read the contents of the directory '".$directory."'<p>");
		#@files = grep readdir(dir); # that gets all files
		#@files = grep {/\w+\..\w+/} readdir(dir); # that gets files only
		@files = grep {/^[^\.]/} readdir(dir); # that gets files that don't start with a dot
		closedir dir;
		return @files;
	}
	
	sub get_recursive_listing { # return a list of all the files in a given folder, including subfolders
		my $start_folder = $_[0]; my $recursive = $_[1]; my $filter = $_[2];
		$start_folder =~ s/[\/\\]$//; # remove a trailing slash
		@file_list = ();
		@folder_list = $start_folder;
		while (@folder_list) {
			$this_folder_path = @folder_list[0];
			@this_folder_items = get_directory_listing($this_folder_path);
			for ($n=0; $n<@this_folder_items; $n++) {
				$this_item_name = @this_folder_items[$n];
				$this_item_path = $this_folder_path."/".$this_item_name;
 				# if the item name contains a dot, we assume it's a file, if not, we assume it's a directory
 				if ($this_item_name !~ /\./) {
 					# we've found a folder; add it to the list of folders to scan
 					if ($recursive) {
						push(@folder_list, $this_item_path);
					}
				} elsif ((!$filter)||($this_item_name =~ /$filter/)) {
					# we've found a file; add it to the list of files to return
					$this_item_path =~ s/$start_folder\///; # make results relative to the start folder, not to the site
					push(@file_list, $this_item_path);
				}
			}
			shift @folder_list; # remove the folder we just scanned from the list
		}
		return @file_list;
	}
	
	sub absolute_path { # convert a reference to a file within our site to a complete filesystem path or absolute URL
		my $path = $_[0]; my $method = $_[1];
		
		if ($method eq "filesystem") {
		
			if ($script_location eq "NA") {
				$output .= show_error("Couldn't determine the filesystem path to this script. Please ensure that either the SCRIPT_FILENAME or PATH_TRANSLATED variable is available on your server.");
			} else {
				$script_location_copy = $script_location;
				$path_copy = $path;
				while ($path_copy =~ /\.\.\//) {
					# this avoids results like contemplate/../templates/default.html
					$script_location_copy =~ s/[^\/]+\/$//; # remove the last directory name from the script location
					$path_copy =~ s/\.\.\///; # remove the first updelim from the filename
				}
				$path = $script_location_copy.$path_copy;
			}

		} else {
		
			if ($request eq "NA") {
				$output .= show_error("Couldn't determine the web path to this script. Please ensure that either the SCRIPT_NAME or REQUEST_URI variable is available on your server.");
			} else {
				$thisserver = $ENV{"HTTP_HOST"};
				$thispath = "";
				if ($URL_rewriting_flag) {
					if ($request =~ /\/([^=]*\/)$URL_rewriting_flag\//i) { $thispath = $1; }
				} else {
					if ($request =~ /\/([^=]*\/)contemplate\//i) { $thispath = $1; }
				}
				
				# to do: how can we read the authentication info from this request so we can pass it on?
				$login = ""; 
				
				$path =~ s/\.\.\//http:\/\/$login$thisserver\/$thispath/i; # replace the first updelim with the scheme and path
				$path =~ s/\.\.\///g; # remove additional updelims
			}
		
		}
		
		return $path;
	}

	sub show_error { # the only purpose for this here is to prevent the script from halting if one of the shared functions throws an error
		my $message = $_[0];
		return "Error: ".$message;
	}
	

# set up some error messages that we can use later

	$check_URL = "<br>Hint: Check your page definition";
	$check_embed = "<br>Hint: Check your embed tag";
	$check_field_label = "<br>Hint: Check your content file and remove any formatting tags between the field marker and the field label";
	

# set up some primary variables

	# try to figure out the path to this script as it appears in the browser, before any URL rewriting
	# in IIS, this probably won't work, because REQUEST_URI will either not exist, or will show the path after rewriting, since rewriting is handled externally from IIS
	$request = "NA";
	if ($ENV{'REQUEST_URI'}) {
		$request = $ENV{'REQUEST_URI'};
		if (index($request, "?") != -1) {
			$request = substr($request, 0, index($request, "?"));
		}
	} else {
		# IIS doesn't always provide the REQUEST_URI variable
		if ($ENV{'SCRIPT_NAME'}) {
			$request = $ENV{'SCRIPT_NAME'};
		}
	}

	# we don't do URL rewriting except in the assembler, so this value should be empty
	# also, we can't tell what flag(s) the assembler is using, but we can set a default here
	$URL_rewriting_flag = "";
	$URL_rewriting_flag_default = "assembled";

	# figure out the full local path of the contemplate directory so we can build an absolute path that works on Windows computers
	# on Unix, this will include any symlinks used to construct the URL, so we can run the script from a symlink to the real contemplate folder, 
	# and paths will be relative to the apparent, rather than the actual, location of assembler.asp
	# on Windows, this will not include virtual directories, but constructing a path from a virtual directory wouldn't work anyway
	# because those aren't available at the file system level
	$script_location = "NA";
	if ($ENV{'SCRIPT_FILENAME'}) {
		$script_location = $ENV{'SCRIPT_FILENAME'};
	} else {
		# Windows doesn't provide the SCRIPT_FILENAME variable
		if ($ENV{'PATH_TRANSLATED'}) {
			$script_location = $ENV{'PATH_TRANSLATED'};
		}
	}
	$script_location =~ s/\\/\//g;
	if (($URL_rewriting_flag)&&($script_location =~ /(.*\/)$URL_rewriting_flag\//i)) {
		$script_location = $1."contemplate/";
	} else {
		$script_location =~ s/(.*\/)[^\/]+/$1/;
	}
	

# get some preference settings from the preferences file

	$preferences = get_file("../data/preferences.ini", "filesystem", 0);
	if ($preferences) { $read_page_definitions_from = "../".get_setting($preferences, "read page definitions from", 0, 1); }
	if (($read_page_definitions_from eq "../")||($read_page_definitions_from eq "")) { $read_page_definitions_from = "../data/pages.txt"; }
	if ($preferences) { $read_files_recursively = get_setting($preferences, "read files recursively", 0, 1); }
	if ($read_files_recursively eq "") { $read_files_recursively = "1"; }
	$read_files_recursively = int($read_files_recursively);


# set up the HTML page
	
	print "
	
		<html>
	
		<head>
		
			<title>Contemplate</title>
			
			<link rel=stylesheet href=../stylesheet.css type=text/css>
		
		</head>
		
		<body bgcolor=#FFFFFF>
				
			<form name=contemplate>
	
	";
	
	
# print a list of available templates by scannning the templates folder
	
	@templates = get_recursive_listing("../../templates", $read_files_recursively, "");
	$templates_count = @templates;
			
	if ($action eq "template") {
	
		print "<p class=heading>Choose a template with which to build your new page</p>\n\n";
		
		print "<p class=body>\n\n";
		
		for ($i=0; $i<$templates_count; $i++) {
		
			$this_template = @templates[$i];
			$this_template_checked = ($stored_template eq $this_template) ? "checked" : "" ;
			print "<input type=radio name=choices onclick=\"parent.set_choice($i,'template','$this_template',0)\" $this_template_checked>\n";
			print "<a href=\"JavaScript:parent.set_choice($i,'template','$this_template',1)\">$this_template</a><br>\n\n";
			
		}
					
		print "</p>\n\n";
	
	}
	
	
# print a list of defined pages by scanning the page definitions file
					
	$pages_file = get_file($read_page_definitions_from, "filesystem", 0);
	@pages = get_page_names($pages_file);
	$pages_count = @pages;
	
	@definitions = ();
									
	if (($action eq "page")||($action eq "edit")||($action eq "delete")) {

		if ($action eq "page") {
			print "<p class=heading>Choose a page on which to base your new page</p>\n\n";
		} elsif ($action eq "edit") {
			print "<p class=heading>Choose the page you want to edit</p>\n\n";
		} else {
			print "<p class=heading>Choose the page you want to delete</p>\n\n";
		}
				
		print "<p class=body>\n\n";
		
		for ($i=0; $i<$pages_count; $i++) {
	
			$this_page = @pages[$i];
			$this_page_checked = ($stored_page eq $this_page) ? "checked" : "" ;
			print "<input type=radio name=choices onclick=\"parent.set_choice($i,'page','$this_page',0)\" $this_page_checked>\n";
			print "<a href=\"JavaScript:parent.set_choice($i,'page','$this_page',1)\">$this_page</a><br>\n\n";
			
			$this_definition = get_setting($pages_file, $this_page, 0, 0);
			$this_definition =~ s/'/\\'/g; # in case the definition contains an apostrophe, which would break the JavaScript array setup code if not escaped
			push(@definitions, $this_definition);
			
		}
					
		print "</p>\n\n";
		
	}
	
	
# set some variables in JavaScript based on the stuff we scanned above
					
	print "<script language=JavaScript>\n\n";
	
	print "parent.page_names = new Array('";
	print join ("', '", @pages);
	print "');\n";
	
	print "parent.page_definitions = new Array('";
	print join ("', '", @definitions);
	print "');\n\n";
			
	print "parent.template_files = new Array('";
	print join ("', '", @templates);
	print "');\n";

	print "</script>\n\n";


# we're done!

	print "</form>";

	print "</body></html>";

	exit;
