#!/usr/bin/perl


# Contemplate Reporter (reporter.cgi) version 1.2.3
# Copyright 1999-2008 Type A Learning Agency (http://www.typea.net)
# This file may be used, modified, or redistributed under the terms
# of the open-source agreement contained in docs/license.txt


# use the CGI.pm module to access CGI functionality

	use CGI;
	$cgi = new CGI;



# define some useful subroutines

	sub get_file_simple {
		my $file = $_[0]; $text = "";

		$file =~ s/[^\w-\.\/\?\&\=\%]//g; # don't read files whose names contain characters that could be used to read remote files, or to hack into the open command (Perl only)
		$file = absolute_path($file, "filesystem");
		open (myfile, $file) or
		$text = "Error: Can't find file $file";
		unless ($text =~ /^Error: /) {
			while (<myfile>) { $text .= "$_"; }
			close (myfile);
		}
		
		# conform all line breaks to avoid cross-platform issues
		$text =~ s/\r\n/\r/g; # change DOS to Mac
		$text =~ s/\n/\r/g; # change Unix to Mac
		
		# remove extra line breaks at the end of the file so all the languages perform splits in the same way
		$text =~ s/(.*[^\r])\r*$/$1/;
		
		return $text;
	}
	
	sub get_setting {
		my $source_text = $_[0]; my $setting = $_[1]; my $allow_override = $_[2]; my $allow_partial = $_[3]; my $value = "";
		
		# allow most preferences to be overridden through the argument string
		$setting_no_spaces = $setting;
		$setting_no_spaces =~ s/[ -]/_/g;
		$settings_used .= "$setting_no_spaces,";

		$override = $cgi->param($setting_no_spaces);
		if (($allow_override)&&($override ne "")) {
			$value = $override;
			# don't allow users to pass code through the URL arguments
			$value =~ s/</&lt;/i;
			$value =~ s/>/&gt;/i;
		} else {
			$source_text = "\r".$source_text;  # this makes it easy to make sure settings start on a new line
			if ($allow_partial) {  # this lets us use more verbose labels in the ini file
				if ($source_text =~ /[^\w]$setting[^:\r]*: ([^\r]*)\r*/i) {
					$value = $1;
				}
			} else {  # this ensures an exact match in the pages file
				if ($source_text =~ /[^\w]$setting: ([^\r]*)\r*/i) {
					$value = $1;
				}
			}
		}
		if ($value eq "0") { $value = 0; } # this ensures that a 0 value will evaluate false in every language
		return $value;
	}
	
	sub absolute_path { # convert a reference to a file within our site to a complete filesystem path or absolute URL
		my $path = $_[0]; my $method = $_[1];
		
		if ($method eq "filesystem") {
		
			if ($script_location eq "NA") {
				$output .= show_error("Couldn't determine the filesystem path to this script. Please ensure that either the SCRIPT_FILENAME or PATH_TRANSLATED variable is available on your server.");
			} else {
				$script_location_copy = $script_location;
				$path_copy = $path;
				while ($path_copy =~ /\.\.\//) {
					# this avoids results like contemplate/../templates/default.html
					$script_location_copy =~ s/[^\/]+\/$//; # remove the last directory name from the script location
					$path_copy =~ s/\.\.\///; # remove the first updelim from the filename
				}
				$path = $script_location_copy.$path_copy;
			}

		} else {
		
			if ($request eq "NA") {
				$output .= show_error("Couldn't determine the web path to this script. Please ensure that either the SCRIPT_NAME or REQUEST_URI variable is available on your server.");
			} else {
				$thisserver = $ENV{"HTTP_HOST"};
				$thispath = "";
				if ($URL_rewriting_flag) {
					if ($request =~ /\/([^=]*\/)$URL_rewriting_flag\//i) { $thispath = $1; }
				} else {
					if ($request =~ /\/([^=]*\/)contemplate\//i) { $thispath = $1; }
				}
				
				# to do: how can we read the authentication info from this request so we can pass it on?
				$login = ""; 
				
				$path =~ s/\.\.\//http:\/\/$login$thisserver\/$thispath/i; # replace the first updelim with the scheme and path
				$path =~ s/\.\.\///g; # remove additional updelims
			}
		
		}
		
		return $path;
	}

	sub random_color {
		# returns a random web-safe color definition
		@colors = ("00","33","66","99","CC","FF");
		$r = @colors[rand(5)]; #rand(5) instead of rand(6) keeps them from getting too light to read
		$g = @colors[rand(5)];
		$b = @colors[rand(5)];
		$color = "\#".$r.$g.$b;
		return $color;		
	}
	
	sub month_display { # converts 2003-05 to May 2003
		my $month = $_[0];
		@month_year = split("-", $month);
		$month_display = @months[int(@month_year[1]) - 1];
		$year_display = @month_year[0];
		return "$month_display $year_display";
	}
	
	sub array_unique {
		my @array = @_; # a list passed in as a parameter becomes separate parameters; this puts them all back into @array again
		@reversed = ();
		@result = ();
		for ($n=0; $n<@array; $n++) {
			$reversed{@array[$n]} = @array[$n];
		}
		@result = keys(%reversed);
		return @result;
	}
	
	sub get_unix_time { # converts an SQL date (yyyy-mm-dd or yyyy-mm-dd hh:mm:ss) to a Unix timestamp for use in PHP datetime functions
		my $date_time = $_[0];
		
		use Time::Local;
		
		@date_time = split(" ", $date_time);
		if (@date_time == 2) {
			$date = @date_time[0];
			$time = @date_time[1];
		} else {
			$date = @date_time[0];
			$time = "12:00:00";
		}
		
		@date = split("-", $date);
		@time = split(":", $time);
		
		my $year = @date[0] - 1900;
		my $month = @date[1] - 1;
		my $day = @date[2];
		
		my $hours = @time[0];
		my $minutes = @time[1];
		my $seconds = @time[2];

		$timestamp = timelocal($seconds, $minutes, $hours, $day, $month, $year);
		return $timestamp;
	}

	sub wordwrap { # this is a substitute for PHP's wordwrap built-in function, with the $cut argument always true
		my $string = $_[0]; my $width = $_[1]; my $delimiter = $_[2];
		$result = "";
		while (length($string) > $width) {
			$result .= substr($string, 0, $width).$delimiter;
			$string = substr($string, $width, length($string) - $width);
		}
		$result .= $string;
		return $result;
	}

	
# set up some primary variables

	# try to figure out the path to this script as it appears in the browser, before any URL rewriting
	# in IIS, this probably won't work, because REQUEST_URI will either not exist, or will show the path after rewriting, since rewriting is handled externally from IIS
	$request = "NA";
	if ($ENV{'REQUEST_URI'}) {
		$request = $ENV{'REQUEST_URI'};
		if (index($request, "?") != -1) {
			$request = substr($request, 0, index($request, "?"));
		}
	} else {
		# IIS doesn't always provide the REQUEST_URI variable
		if ($ENV{'SCRIPT_NAME'}) {
			$request = $ENV{'SCRIPT_NAME'};
		}
	}

	# we don't do URL rewriting except in the assembler, so this value should be empty
	# also, we can't tell what flag(s) the assembler is using, but we can set a default here
	$URL_rewriting_flag = "";
	$URL_rewriting_flag_default = "assembled";

	# figure out the full local path of the contemplate directory so we can build an absolute path that works on Windows computers
	# on Unix, this will include any symlinks used to construct the URL, so we can run the script from a symlink to the real contemplate folder, 
	# and paths will be relative to the apparent, rather than the actual, location of assembler.asp
	# on Windows, this will not include virtual directories, but constructing a path from a virtual directory wouldn't work anyway
	# because those aren't available at the file system level
	$script_location = "NA";
	if ($ENV{'SCRIPT_FILENAME'}) {
		$script_location = $ENV{'SCRIPT_FILENAME'};
	} else {
		# Windows doesn't provide the SCRIPT_FILENAME variable
		if ($ENV{'PATH_TRANSLATED'}) {
			$script_location = $ENV{'PATH_TRANSLATED'};
		}
	}
	$script_location =~ s/\\/\//g;
	if (($URL_rewriting_flag)&&($script_location =~ /(.*\/)$URL_rewriting_flag\//i)) {
		$script_location = $1."contemplate/";
	} else {
		$script_location =~ s/(.*\/)[^\/]+/$1/;
	}

	$month = $cgi->param('month');
	$section = $cgi->param('section');
	$clear = $cgi->param('clear');
	$inspector_key = $cgi->param('inspector_key');
	$inspector_value = $cgi->param('inspector_value');
	$back_label = $cgi->param('back_label');
	$realtime = $cgi->param('realtime');
		
	$output = "";
	$button = "";
	$settings_used = "";
	$page_views_greatest = 1;
	$blank_marker = "--";
	$graph_width = 140;
	$my_host = $ENV{'HTTP_HOST'};
	$my_domain = ($my_host =~ /www\./) ? substr($my_host, 3, length($my_host)) : $my_host ;

	$show_performance = 0;
	$performance_times = "";
	$last_performance_time = time;
	$total_performance_time = 0;

	$data_directory = "../data/reporter/";
	$data_file = $data_directory.$month.".txt";
	$error_file = $data_directory.$month."-errors.txt";

	@columns = ("dates", "visitors", "IP_addresses", "hosts", "pages", "user_agents", "referrers");
	@columns_errors = ("dates", "visitors", "IP_addresses", "hosts", "pages", "user_agents", "referrers", "message");

	@sections = ("dates", "pages", "referrers", "entry_pages", "exit_pages", "visitors", "hosts", "countries", "IP_addresses", "user_agents");
	@sections_display = ("dates", "pages", "referrers", "entry pages", "exit pages", "visitors", "hosts", "countries", "IP addresses", "user agents");
	for ($i=0; $i<@sections; $i++) {
		@{@sections[$i]} = ();
	}	

	@months = ("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");

	%HTTP_error_codes = (
		"404", "File not found",
		"403", "Forbidden",
		"401", "Not authorized"
	);

	%country_codes = (
		"com", "US Commercial",
		"edu", "US Educational",
		"org", "US Organization",
		"net", "US Network",
		"gov", "US Government",
		"mil", "US Military",
		"ac", "Ascension Island",
		"ad", "Andorra",
		"ae", "United Arab Emirates",
		"af", "Afghanistan",
		"ag", "Antigua and Barbuda",
		"ai", "Anguilla",
		"al", "Albania",
		"am", "Armenia",
		"an", "Netherland Antilles",
		"ao", "Angola",
		"aq", "Antarctica",
		"ar", "Argentina",
		"as", "American Samoa",
		"at", "Austria",
		"au", "Australia",
		"aw", "Aruba",
		"ax", "Aland Islands",
		"az", "Azerbaijan",
		"ba", "Bosnia and Herzegowina",
		"bb", "Barbados",
		"bd", "Bangladesh",
		"be", "Belgium",
		"bf", "Burkina Faso",
		"bg", "Bulgaria",
		"bh", "Bahrain",
		"bi", "Burundi",
		"bj", "Benin",
		"bm", "Bermuda",
		"bn", "Brunei Darussalam",
		"bo", "Bolivia",
		"br", "Brazil",
		"bs", "Bahamas",
		"bt", "Bhutan",
		"bv", "Bouvet Island",
		"bw", "Botswana",
		"by", "Belarus",
		"bz", "Belize",
		"ca", "Canada",
		"cc", "Cocos (Keeling) Islands",
		"cd", "Congo",
		"cf", "Central African Republic",
		"cg", "Congo",
		"ch", "Switzerland",
		"ci", "Cte d'Ivoire",
		"ck", "Cook Islands",
		"cl", "Chile",
		"cm", "Cameroon",
		"cn", "China",
		"co", "Columbia",
		"cr", "Costa Rica",
		"cs", "Czechoslovakia",
		"cu", "Cuba",
		"cv", "Cape Verde",
		"cx", "Christmas Island",
		"cy", "Cyprus",
		"cz", "Czech Republic",
		"de", "Germany",
		"dj", "Djibouti",
		"dk", "Denmark",
		"dm", "Dominica",
		"do", "Dominican Republic",
		"dz", "Algeria",
		"ec", "Ecuador",
		"ee", "Estonia",
		"eg", "Egypt",
		"eh", "Western Sahara",
		"er", "Eritrea",
		"es", "Spain",
		"et", "Ethiopia",
		"eu", "European Union",
		"fi", "Finland",
		"fj", "Fiji",
		"fk", "Falkland Islands",
		"fm", "Micronesia",
		"fo", "Faroe Islands",
		"fr", "France",
		"ga", "Gabon",
		"gb", "United Kingdom",
		"gd", "Grenada",
		"ge", "Georgia",
		"gf", "French Guiana",
		"gg", "Guernsey",
		"gh", "Ghana",
		"gi", "Gibraltar",
		"gl", "Greenland",
		"gm", "Gambia",
		"gn", "Guinea",
		"gp", "Guadelope",
		"gq", "Equatorial Guinea",
		"gr", "Greece",
		"gs", "S. Georgia/S. Sandwich Is.",
		"gt", "Guatemala",
		"gu", "Guam",
		"gw", "Guinea-Bissau",
		"gy", "Guyana",
		"hk", "Hong Kong",
		"hm", "Heard and McDonald Islands",
		"hn", "Honduras",
		"hr", "Croatia",
		"ht", "Haiti",
		"hu", "Hungary",
		"id", "Indonesia",
		"ie", "Ireland",
		"il", "Israel",
		"im", "Isle of Man",
		"in", "India",
		"io", "British Indian Ocean Territory",
		"iq", "Iraq",
		"ir", "Iran",
		"is", "Iceland",
		"it", "Italy",
		"je", "Jersey",
		"jm", "Jamaica",
		"jo", "Jordan",
		"jp", "Japan",
		"ke", "Kenya",
		"kg", "Kyrgystan",
		"kh", "Cambodia",
		"ki", "Kiribati",
		"km", "Comoros",
		"kn", "Saint Kitts and Nevis",
		"kp", "Korea, People's Republic of",
		"kr", "Korea, Republic of",
		"kw", "Kuwait",
		"ky", "Cayman Islands",
		"kz", "Kazakhstan",
		"la", "Lao Democratic Republic",
		"lb", "Lebanon",
		"lc", "Saint Lucia",
		"li", "Liechtenstein",
		"lk", "Sri Lanka",
		"lr", "Liberia",
		"ls", "Lethoso",
		"lt", "Lithuania",
		"lu", "Luxembourg",
		"lv", "Latvia",
		"ly", "Libyan Arab Jamahiriya",
		"ma", "Morocco",
		"mc", "Monaco",
		"md", "Moldova",
		"mg", "Madagascar",
		"mh", "Marshall Islands",
		"mk", "Macedonia",
		"ml", "Mali",
		"mm", "Myanmar",
		"mn", "Mongolia",
		"mo", "Macau",
		"mp", "Northern Mariana Islands",
		"mq", "Martinique",
		"mr", "Mauritania",
		"ms", "Montserrat",
		"mt", "Malta",
		"mu", "Mauritius",
		"mv", "Maldives",
		"mw", "Malawi",
		"mx", "Mexico",
		"my", "Malaysia",
		"mz", "Mozambique",
		"na", "Namibia",
		"nc", "New Caledonia",
		"ne", "Niger",
		"nf", "Norfolk Island",
		"ng", "Nigeria",
		"ni", "Nicaragua",
		"nl", "The Netherlands",
		"no", "Norway",
		"np", "Nepal",
		"nr", "Nauru",
		"nu", "Niue",
		"nz", "New Zealand",
		"om", "Oman",
		"pa", "Panama",
		"pe", "Peru",
		"pf", "French Polynesia",
		"pg", "Papua New Guinea",
		"ph", "Philippines",
		"pk", "Pakistan",
		"pl", "Poland",
		"pm", "St. Pierre and Miquelon",
		"pn", "Pitcairn",
		"pr", "Puerto Rico",
		"ps", "Palestinian Territories",
		"pt", "Portugal",
		"pw", "Palau",
		"py", "Paraguay",
		"qa", "Qatar",
		"re", "Reunion",
		"ro", "Romania",
		"ru", "Russia",
		"rw", "Rwanda",
		"sa", "Saudi Arabia",
		"sb", "Solomon Islands",
		"sc", "Seychelles",
		"sd", "Sudan",
		"se", "Sweden",
		"sg", "Singapore",
		"sh", "Saint Helena",
		"si", "Slovenia",
		"sj", "Svalbard/Jan Mayen Islands",
		"sk", "Slovakia",
		"sl", "Sierra Leone",
		"sm", "San Marino",
		"sn", "Senegal",
		"so", "Somalia",
		"sr", "Surinam",
		"st", "Sao Tome and Principe",
		"sv", "El Salvador",
		"sy", "Syrian Arab Republic",
		"sz", "Swaziland",
		"tc", "The Turks & Caicos Islands",
		"td", "Chad",
		"tf", "French Southern Territories",
		"tg", "Togo",
		"th", "Thailand",
		"tj", "Tajikistan",
		"tk", "Tokelau",
		"tl", "Timor-Leste",
		"tm", "Turkmenistan",
		"tn", "Tunesia",
		"to", "Tonga",
		"tp", "East Timor",
		"tr", "Turkey",
		"tt", "Trinidad and Tobago",
		"tv", "Tuvalu",
		"tw", "Taiwan",
		"tz", "Tanzania",
		"ua", "Ukraine",
		"ug", "Uganda",
		"uk", "United Kingdom",
		"um", "U.S. Minor Outlying Islands",
		"us", "United States",
		"uy", "Uruguay",
		"uz", "Uzbekistan",
		"va", "Vatican City State",
		"vc", "Saint Vincent & Grenadines",
		"ve", "Venezuela",
		"vg", "British Virgin Islands",
		"vi", "U.S. Virgin Islands",
		"vn", "Vietnam",
		"vu", "Vanuatu",
		"wf", "Wallis and Futuna Islands",
		"ws", "Samoa",
		"ye", "Yemen",
		"yt", "Mayotte",
		"yu", "Yugoslavia",
		"za", "South Africa",
		"zm", "Zambia",
		"zw", "Zimbabwe"
	);
	
	%country_codes_reversed = reverse %country_codes;
	
	
# read in a preferences (.ini) file if one exists, and show which preferences may be overridden from the URL string

	$preferences = get_file_simple("../data/preferences.ini", "filesystem");
	if ($preferences) {
		#$URL_rewriting_flag = get_setting($preferences, "URL rewriting flag", 1, 1);
		$top_n = get_setting($preferences, "top n", 1, 1);
		$graph_image = get_setting($preferences, "graph image", 1, 1);
		$graph_color = get_setting($preferences, "graph color", 1, 1);
		$update_interval = get_setting($preferences, "real-time update interval", 1, 1);
		$maximum_column_width = get_setting($preferences, "column width", 1, 1);
		$show_inspector_rollover = get_setting($preferences, "inspector rollover", 1, 1);
		$one_big_page = get_setting($preferences, "one big page", 1, 1);
		$exclude_from_hosts = get_setting($preferences, "exclude traffic", 1, 1);
		$track_visitors = get_setting($preferences, "track visitors", 1, 1);
		$visit_expires = get_setting($preferences, "visit expires", 1, 1);
	} else {
		#$URL_rewriting_flag = "";
		$top_n = "";
		$graph_color = "";
		$graph_image = "";
		$update_interval = "";
		$maximum_column_width = "";
		$show_inspector_rollover = "";
		$one_big_page = "";
		$exclude_from_hosts = "";
		$track_visitors = "";
		$visit_expires = "";
	}
	

# do a bit of processing on the settings, and set some defaults in case preferences.ini or individual settings are absent

	#if ($URL_rewriting_flag eq "") { $URL_rewriting_flag = "assembled"; }

	if ($top_n eq "") { $top_n = 5; }

	if ($graph_color eq "") { $graph_color = "#0000FF"; }
	$graph_color =~ s/#//; # Flash doesn't use # to denote a hex value

	if ($graph_image eq "") { $graph_image = "../graphics/logo.jpg"; }
	$graph_image = "../".$graph_image; # make the path relative to the Flash file, which is in a subdirectory of the Contemplate directory
	$graph_image =~ s/\.\.\/\.\.\//%252E%252E%252F%252E%252E%252F/; # double-encode the up delimiter to prevent Mac IE from decoding and then rewriting it
	
	if ($update_interval eq "") { $update_interval = 30; }

	if ($maximum_column_width eq "") { $maximum_column_width = 64; }

	# this is eye candy that probably slows down page rendering somewhat
	if ($show_inspector_rollover eq "") { $show_inspector_rollover = 1; }
	$show_inspector_rollover = int($show_inspector_rollover);
	
	# this shows the full data for each section on the same page as the top $top_n
	if ($one_big_page eq "") { $one_big_page = 0; }
	$one_big_page = int($one_big_page);
	
	if ($exclude_from_hosts) {
		$exclude_from_hosts =~ s/, /,/g;
		@exclude_from_hosts = split(",", $exclude_from_hosts);
		$exclude_from_hosts_count = @exclude_from_hosts;
	}

	if ($track_visitors eq "") { $track_visitors = "0"; }
	$track_visitors = int($track_visitors);
	if (!$track_visitors) {
		# remove the visitors elements from our arrays if we're not using them
		splice(@columns, 1, 1);
		splice(@columns_errors, 1, 1);
		splice(@sections, 3, 3);
		splice(@sections_display, 3, 3);
	}

	if ($visit_expires eq "") { $visit_expires = 30; }


# if requested, delete the error log

	if ($clear) {
		unlink $error_file;
	}
		

# write a generic header to prepare for text output (PHP and ASP normally don't need this)

	$header = $cgi->header;
	print $header;
	

# if called from the real-time Flash viewer, provide some data for that

	if ($realtime) {
	
		# reinitialize $data_file to use the current month rather than looking in the URL string for a month
		($second, $minute, $hour, $day, $month, $year, $weekday, $yearday, $dst_flag) = localtime(time);
		$year += 1900;
		$month = sprintf("%02d", ($month + 1) % 100); 
		$data_file = "$year-$month.txt";
		$data_file_text = $data_file;
		$data_file = $data_directory.$data_file;

		# open the log file
		$text = get_file_simple($data_file);
		
		# create an array from the lines in the file
		@text = split("\r", $text);
		
		# loop through the array and count the number of lines that contain the appropriate number of columns	
		$page_views_total = 0;
		for ($line_index=0; $line_index < @text; $line_index++) {
			@next = split("\t", @text[$line_index], 500); # the last argument, in part, tells the split function not to strip trailing empty elements
			if (@next == @columns) {
				$page_views_total++;
			}
		}		

		# grab just the last ten lines to pass back to the Flash viewer
		if (@text > 10) {
			@tail = splice(@text, -10);
		}
		
		# process those lines a bit before passing them on
		for ($line_index=0; $line_index < @tail; $line_index++) {
			@next = split("\t", $tail[$line_index], 500);

			# add the country code to each line	
			$host = ($track_visitors) ? @next[3] : @next[2] ;
			if ($host =~ /\.([^\.\d]+)$/) {
				push(@next, @country_codes{$1});
			} else {
				push(@next, $blank_marker);
			}

			# remove the visitor ID so the Flash viewer doesn't have to worry about it
			if ($track_visitors) {
				splice(@next, 1, 1);
			}
			
			$tail[$line_index] = join("\t", @next);
		}
				
		# finally, print the results out for the Flash viewer to read
		$tail = join("\n", @tail);
		$tail =~ s/ /%20/g;
		$tail =~ s/\=/%3D/g;
		$tail =~ s/\&/%26/g;
		$tail =~ s/\?/%3F/g;
		$tail =~ s/\+/%2B/g;
		$tail =~ s/\(/%28/g;
		$tail =~ s/\)/%29/g;
		$tail =~ s/\;/%3B/g;
		$tail =~ s/\[/%5B/g;
		$tail =~ s/\]/%5D/g;
		print "hits=$page_views_total&data_file=$data_file_text&tail=$tail";
		
		exit;
	
	}


# set up the HTML page
	
	print "
	
		<html>
	
		<head>
		
			<title>Contemplate(tm) Web Templating System</title>
			
			<link rel=stylesheet href=../stylesheet.css type=text/css>

		</head>
		
		<body bgcolor=#FFFFFF>
				
			<p class=title>Contemplate</p>
			<p class=body><a href=../index.html>Utilities menu</a> &gt; Reporter</p>
	
			<div class=body>
	
	";
	
	if ($show_performance) {
		$this_performance_time = time;
		$performance_times .= "&nbsp;&nbsp;script and page setup - ".($this_performance_time - $last_performance_time)." seconds<br>\r";
		$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
		$last_performance_time = $this_performance_time;
	}


# if an inspect key is provided, focus on a cross-section of the month's report

	if ($inspector_key) {

	
	# don't display the entry and exit page values on the inspector page

		splice(@sections, 3, 2);
		splice(@sections_display, 3, 2);
	

	# read the data file for the current month and save the lines that match the inspect value
	
		$text = get_file_simple($data_file);
		if ($text =~ /(visitors not tracked)/) {
			# don't display the visitor values if we encounter a file that predates that functionality
			splice(@sections, 3, 2);
			splice(@sections_display, 3, 2);
		}
		@text = split("\r", $text);
		@lines = ();

		# first, make a list of the lines we want to display
		# we'll do this differently for entry and exit pages because we want to show them with their corresponding visit pages
		if (($inspector_key ne "entry_pages")&&($inspector_key ne "exit_pages")) {

			# it's nice for users to pass in the column name as the inspect key, but the column index is more useful to us
			if ($inspector_key eq "countries") {
				# countries is really just a part of the host value
				$inspector_key_index = 3;
			} else {
				for ($i=0; $i<@columns; $i++) {
					if ($inspector_key eq @columns[$i]) { $inspector_key_index = $i; }
				}
			}
		
			for ($line_index=0; $line_index<@text; $line_index++) {
				@next = split("\t", @text[$line_index], 500);
				if (@next == @columns) { # make sure the line has all the elements we're expecting it to have
				
					# dates have to start with value, country codes have to end with value, all others have to equal value
					$match_end = ($inspector_key eq "dates") ? ".*" : "" ;
					$match_start = ($inspector_key eq "countries") ? ".*" : "" ;
					$match_value = $match_start.$inspector_value.$match_end;
					if (@next[$inspector_key_index] =~ /^$match_value$/) {
						push(@lines, $line_index);
					}
					
				}
			}
		
		} else {
		
			# we're going to make a $visits array that's an array of dates, each of which is in turn an array of line numbers from the data file
			@visits = ();

			for ($line_index=0; $line_index<@text; $line_index++) {
				@next = split("\t", @text[$line_index], 500);
				if (@next == @columns) { # make sure the line has all the elements we're expecting it to have
				
					# extract the date and visitor values out of the log entry
					$this_dates = $next[0];
					$this_visitors = $next[1];

					# don't track this at all if the visitor ID is missing (e.g. the client didn't accept the cookie)
					if (!$this_visitors) {
						$page_views--;
					} else {
					
						if (!$visits{$this_visitors}) {
							# if we haven't encountered this visitor yet, add a new visit entry with this date and line number
							@visits{$this_visitors} = ();
							@{$visits{$this_visitors}{$this_dates}} = ();
							push(@{$visits{$this_visitors}{$this_dates}}, $line_index);
							#print "line $line_index: adding visitor $this_visitors (".join(",", @{$visits{$this_visitors}{$this_dates}}).")<br>"; # debug
						} else {
							# if we have encountered this visitor already, check the last page view to see if this is part of the same visit
							@visit_dates = sort(keys(%{$visits{$this_visitors}})); # we need to sort this because the hash doesn't necessarily stay in order and this ensures that we get the last date
							$last_date = @visit_dates[@visit_dates - 1];
							$interval = (get_unix_time($this_dates) - get_unix_time($last_date));
							if ($interval < ($visit_expires * 60)) {
								# if the last page view was less than $visit_expires minutes ago, add this line number to the latest visit array
								push(@{$visits{$this_visitors}{$last_date}}, $line_index);
								#print "line $line_index: updating visit $last_date (".join(",", @{$visits{$this_visitors}{$last_date}}).")<br>"; # debug
							} else {	
								# otherwise, make a new visit array for this line number
								@{$visits{$this_visitors}{$this_dates}} = ();
								push(@{$visits{$this_visitors}{$this_dates}}, $line_index);
								#print "line $line_index: adding visit $this_dates (".join(",", @{$visits{$this_visitors}{$this_dates}}).")<br>"; # debug
							}
						}
					
					}

				}
			}

			# now filter and sort our multidimensional $visits array and flatten it into a one-dimensional $lines array
			@visits_flattened = ();
			@lines = ();
			foreach $visitor(keys(%visits)) {
				foreach $date(keys(%{$visits{$visitor}})) {
					$entry_line_index = @{$visits{$visitor}{$date}}[0];
					$exit_line_index = @{$visits{$visitor}{$date}}[@{$visits{$visitor}{$date}} - 1];
					@entry_line = split("\t", @text[$entry_line_index]);
					@exit_line = split("\t", @text[$exit_line_index]);
					$entry_page = @entry_line[4];
					$exit_page = @exit_line[4];
					if ( (($inspector_key eq "entry_pages")&&($inspector_value eq $entry_page)) || (($inspector_key eq "exit_pages")&&($inspector_value eq $exit_page)) ) {
						@{$visits_flattened{"$date $visitor"}} = @{$visits{$visitor}{$date}};
					}
				}
			}
			@visits_flattened_keys = sort(keys(%visits_flattened));
			foreach $visit(@visits_flattened_keys) {
				foreach $line_index(@{$visits_flattened{$visit}}) {
					push(@lines, $line_index);
				}
			}
					
		}
		
		# then, loop through the array of line indexes we created and prepare the selected lines for display
		@lines_prepared = ();
		for ($line_index=0; $line_index<@lines; $line_index++) {
			@next = split("\t", @text[@lines[$line_index]], 500);
			#print "displaying line ".$lines[$line_index]."<br>"; # debug

			# extract the values out of the log entry
			for ($i=0; $i < @columns; $i++) {
				${"this_".@columns[$i]} = (@next[$i] ne "") ? @next[$i] : $blank_marker ;
			}				
			
			# check the host and skip this line if it matches our exclude hosts
			$include_this_host = 1;
			if (@exclude_from_hosts) {
				for ($i=0; $i < $exclude_from_hosts_count; $i++) {
					if ($this_hosts =~ /@exclude_from_hosts[$i]/i) { $include_this_host = 0; }
				}
			}

			if ($include_this_host) {
				# clean up some of the values
				# - display visitor ID as a user agent and timestamp if the timestamp is present
				if ($this_visitors =~ /\w+-(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)/) {
					$this_visitors_display = $1."-".$2."-".$3." ".$4.":".$5.":".$6;
					$this_user_agents_display = $this_user_agents;
					if ($this_user_agents_display =~ /.+\)(.+)/) {
						$this_user_agents_display = $1; # if there is a parenthesis, just take the text after the parenthesis
					}
					$this_user_agents_display =~ s/compatible; //;
					$this_user_agents_display = (length($this_user_agents_display) > ($maximum_column_width - 22)) ? substr($this_user_agents_display,0,$maximum_column_width - 25)."..." : $this_user_agents_display ;
					$this_visitors = $this_user_agents_display." (".$this_visitors_display.")";
				}
				# - convert TLD to country code
				if ($this_hosts =~ /\.([^\.\d]+)$/) {
					$TLD = lc($1);
					$this_countries = @country_codes{$TLD};
				} else {
					$this_countries = $blank_marker;
				}
					
				# then add each value to the appropriate array
				@this_line_prepared = ();
				for ($i=0; $i < @sections; $i++) {
					$this_line_prepared[$i] = ${"this_".@sections[$i]};
				}
				push(@lines_prepared, @this_line_prepared);
			}
			
		}
			
					
	# print the matching lines in a table
	
		$output .= "<p class=heading><a name=top>Traffic for ".month_display($month)."</a></p>\r";
		
		$inspector_key_singular = substr($inspector_key, 0, -1);
		if ($inspector_key_singular eq "IP addresse") { $inspector_key_singular = substr($inspector_key_singular, 0, -1); }
		if ($inspector_key_singular eq "countrie") { $inspector_key_singular = "country"; }
		$output .= "<b>Where $inspector_key_singular contains $inspector_value</b><br>\r";
		$output .= "<b>Total page views: ".@lines_prepared."</b><p>\r\r";
		
		$output .= "&lt;&ndash; <a href=JavaScript:history.back()>$back_label</a><p>\r\r";

		$output .= "<table cellpadding=2 cellspacing=0 border=0>\r\r";
		
		$output .= "<tr bgcolor=#CCCCCC>\r";
		for ($i=0; $i < @sections; $i++) {
			$output .= "<td class=subheading>".ucfirst(@sections[$i])."</td>\r";
		}
		$output .= "</tr>\r\r";
	
		for ($line_index=0; $line_index < @lines_prepared; $line_index++) {	
			# draw a dividing line if we're about to start displaying a new visit (for entry pages and exit pages only)
			if (($inspector_key eq "visitors")||($inspector_key eq "entry_pages")||($inspector_key eq "exit_pages")) {
				$this_dates = @lines_prepared[$line_index][0];
				$this_visitors = @lines_prepared[$line_index][3];
				if ($line_index > 0) {
					if (($this_visitors ne $last_visitor)||(get_unix_time($this_dates) - get_unix_time($last_date)) > ($visit_expires * 60)) {
						$output .= "<tr valign=top><td colspan=".@sections."><hr></td></tr>\r";
					}
				}
				$last_date = $this_dates;
				$last_visitor = $this_visitors;
			}

			$output .= "<tr valign=top>\r";
			for ($i=0; $i < @sections; $i++) {
				$key = @lines_prepared[$line_index][$i];

				# break the line after a predetermined number of characters
				$key_complete = $key;
				$key = (length($key) > $maximum_column_width) ? wordwrap($key, $maximum_column_width, "<br>", 1) : $key ;
				
				# make the page items into a link
				if (@sections[$i] =~ /pages/) {
					$key = ($key ne $blank_marker) ? "<a href=../assembler.cgi?page=$key_complete target='_new'>$key</a>" : $key ;
				}
				
				# make the IP address items into a link
				if (@sections[$i] =~ /IP_addresses/) {
					$key = ($key ne $blank_marker) ? "<a href=http://ws.arin.net/cgi-bin/whois.pl?queryinput=$key_complete target='_new'>$key</a>" : $key ;
				}
				
				# make the referrer items into a link
				if (@sections[$i] =~ /referrers/) {
					$key = ($key ne $blank_marker) ? "<a href=$key_complete target='_new'>$key</a>" : $key ;
				}

				$output .= "<td class=body>\r$key\r</td>\r";
			}
			$output .= "</tr>\r\r";
		}
		
		$output .= "</table><p>\r\r";
	
		$output .= "&lt;&ndash; <a href=JavaScript:history.back()>$back_label</a><p>\r\r";
		
	}


# if a month is specified, display that month's report

	elsif ($month) {


	# read the data file
	
		$text = get_file_simple($data_file);
		if ($text =~ /(visitors not tracked)/) {
			# don't display the entry page, exit page, or visitor values if we encounter a file that predates that functionality
			splice(@sections, 3, 3);
			splice(@sections_display, 3, 3);
		}

		if ($show_performance) {
			$this_performance_time = time;
			$performance_times .= "&nbsp;&nbsp;load data file - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
			$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
			$last_performance_time = $this_performance_time;
		}


	# split the data into lines and columns and create associative arrays

		$page_views_total = 0; # a simple count will do here
		@visits = (); # but here we have to remember each visit to avoid duplicates
		@exit_pages_maybe = ();
		@text = split("\r", $text);
		if ($show_performance) {
			$this_performance_time = time;
			$performance_times .= "&nbsp;&nbsp;split data file - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
			$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
			$last_performance_time = $this_performance_time;
		}
		for ($line_index=0; $line_index<@text; $line_index++) {
			@next = split("\t", @text[$line_index], 500);
			if (@next == @columns) { # make sure the line has all the elements we're expecting it to have
			
				# extract the values out of the log entry
				for ($i=0; $i < @columns; $i++) {
					${"this_".@columns[$i]} = (@next[$i] ne "") ? @next[$i] : $blank_marker ;
				}				
				
				# check the host and skip this line if it matches our exclude hosts
				$include_this_host = 1;
				if (@exclude_from_hosts) {
					for ($i=0; $i < $exclude_from_hosts_count; $i++) {
						if ($this_hosts =~ /@exclude_from_hosts[$i]/i) { $include_this_host = 0; }
					}
				}

				if ($include_this_host) {
					# clean up some of the values
					# - ignore times to group page views by day
					$this_dates = substr($this_dates, 0, index($this_dates, " "));
					# - display visitor ID as a user agent and timestamp if the timestamp is present
					if ($this_visitors =~ /\w+-(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)/) {
						$this_visitors_display = $1."-".$2."-".$3." ".$4.":".$5.":".$6;
						$this_user_agents_display = $this_user_agents;
						if ($this_user_agents_display =~ /.+\)(.+)/) {
							$this_user_agents_display = $1; # if there is a parenthesis, just take the text after the parenthesis
						}
						$this_user_agents_display =~ s/compatible; //;
						$this_user_agents_display = (length($this_user_agents_display) > ($maximum_column_width - 22)) ? substr($this_user_agents_display,0,$maximum_column_width - 25)."..." : $this_user_agents_display ;
						$this_visitors = $this_user_agents_display." (".$this_visitors_display.") ".$this_visitors; # we're adding the original visitors value temporarily so that we can grab it later to build the inspector link
					}
					# - convert TLD to country code
					if ($this_hosts =~ /\.([^\.\d]+)$/) {
						$TLD = lc($1);
						$this_countries = @country_codes{$TLD};
					} else {
						$this_countries = $blank_marker;
					}
					# - filter self-referrers
					if (($this_referrers =~ /$my_domain/)||(length($this_referrers) < 3)) { $this_referrers = $blank_marker; }
						
					# look for entry and exit pages, which requires keeping track of visits (page views by a visitor with no more than a $visit_expires minutes delay between them)
					if ($track_visitors) {
						$this_entry_pages = "";
						$this_exit_pages = "";
						if (!$visits{$this_visitors}) {
							# add a new visit entry with this time
							$visits{$this_visitors} = $this_dates; # we're identifying visits by the visitor ID and date
							# set this page as an entry page
							$this_entry_pages = $this_pages;
							# also hold onto it as a possible exit page (we won't know until we reach the end of the file whether it's really the exit page)
							$exit_pages_maybe{$this_visitors} = $this_pages;
						} else {
							$last_dates = $visits{$this_visitors};
							$interval = (get_unix_time($this_dates) - get_unix_time($last_dates));
							if ($interval < ($visit_expires * 60)) {
								# if the last page view was less than $visit_expires minutes ago, update the visit entry with the time of this latest page view
								$visits{$this_visitors} = $this_dates;
								# also update the possible exit page for this visit
								$exit_pages_maybe{$this_visitors} = $this_pages;
							} else {	
								# otherwise, tag the old visit entry with the end time for that visit and add a new visit entry with this new time
								$visits{"$this_visitors $last_dates"} = $last_dates;
								$visits{$this_visitors} = $this_dates;
								# set this page as an entry page
								$this_entry_pages = $this_pages;
								# move the last page for this visitor from the maybe list to the definite list, then add this page to the maybe list to start the cycle again
								$this_exit_pages = $exit_pages_maybe{$this_visitors};
								$exit_pages_maybe{$this_visitors} = $this_pages;
							}
						}
					}

					# then add each value to the appropriate array
					for ($i=0; $i < @sections; $i++) {
						if (${"this_".@sections[$i]}) { # e.g. if ($this_dates)
							if (exists(${@sections[$i]}{${"this_".@sections[$i]}})) { # e.g. if (exists($dates["2003-05"]))
								${@sections[$i]}{${"this_".@sections[$i]}}++;
								$this_page_views = ${@sections[$i]}{${"this_".@sections[$i]}};
								if ($this_page_views > $page_views_greatest) { $page_views_greatest = $this_page_views; }
							} else {
								%{@sections[$i]} = (%{@sections[$i]}, ${"this_".@sections[$i]} => 1);
							}
						}
					}
						
					$page_views_total++;
				}
			
			}
		}  
		
		# now that we're at the end of the file, any remaining pages in the maybe list are definitely exit pages
		foreach $this_exit_pages(values(%exit_pages_maybe)) {
			if ($this_exit_pages) {
				if (exists($exit_pages{$this_exit_pages})) {
					$exit_pages{$this_exit_pages}++;
					$this_page_views = $exit_pages{$this_exit_pages};
					if ($this_page_views > $page_views_greatest) { $page_views_greatest = $this_page_views; }
				} else {
					$exit_pages{$this_exit_pages} = 1;
				}
			}
		}
						
		if ($show_performance) {
			$this_performance_time = time;
			$performance_times .= "&nbsp;&nbsp;process data file - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
			$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
			$last_performance_time = $this_performance_time;
		}

	
	# print the reports in a table
	
		$output .= "<p class=heading><a name=top>Traffic for ".month_display($month)."</a></p>\r\r";
		
		$output .= "<p class=subheading>Total page views: $page_views_total</p><br>\r\r";
		
		$output .= "<p class=body>\r";
		$output .= "<a href=reporter.cgi>Months</a> &gt; \r";
		if (($section)&&(!$one_big_page)) {
			$output .= "<a href=reporter.cgi?month=$month>Top ".$top_n."'s</a> &gt; \r";
		} else {
			$output .= "Top ".$top_n."'s &gt; \r";
		}
		$output .= "<font color=#0000FF>All</font> ";
		for ($i=0; $i < @sections; $i++) {
			$link = ($one_big_page) ? "\#@sections[$i]" : "reporter.cgi?month=$month&section=@sections[$i]" ;
			if ($section eq @sections[$i]) {
				$output .= ucfirst(@sections_display[$i]).", ";
			} else {
				$output .= "<a href=$link>".ucfirst(@sections_display[$i])."</a>, ";
			}
		}
		if ($section eq "errors_HTTP") {
			$output .= "HTTP errors, ";
		} else {
			if ($one_big_page) {
				$output .= "<a href=\#errors_HTTP>HTTP errors</a>, ";
			} else {
				$output .= "<a href=reporter.cgi?month=$month&section=errors_HTTP>HTTP errors</a>, ";
			}
		}
		if ($section eq "errors_assembler") {
			$output .= "Assembler errors<p>";
		} else {
			if ($one_big_page) {
				$output .= "<a href=\#errors_assembler>Assembler errors</a>\r";
			} else {
				$output .= "<a href=reporter.cgi?month=$month&section=errors_assembler>Assembler errors</a>\r";
			}
		}
		$output .= "</p>\r\r";
		
		$inspector_button_ID = 0;
			
		$output .= "<table cellpadding=2 cellspacing=0 border=0>\r\r";
		
		if ($show_performance) {
			$this_performance_time = time;
			$performance_times .= "&nbsp;&nbsp;set up section links - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
			$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
			$last_performance_time = $this_performance_time;
		}

		$passes = 2;
		for ($h=0; $h < $passes; $h++) { # print all the tables twice, the first time showing just the top $top_n ...
		
			# ... but if we're splitting the sections up, skip the first or second pass
			if (!$one_big_page) {
				if (!$section) {
					$passes = 1; # only do the first part
				} else {
					$h = 1; # only do the second part
				}
			}
			
			for ($i=0; $i < @sections; $i++) {
				
				# skip unless we're displaying this section, or no section is specified
				if ((!$section)||($section eq @sections[$i])) {
			
					if ($h == 0) {
						$label = "Top $top_n @sections_display[$i]";
						$button_link = ($one_big_page) ? "\#@sections[$i]" : "reporter.cgi?month=$month&section=@sections[$i]" ;
						$button = "<a href=$button_link>All @sections_display[$i]</a>";
						$space_after = 18;
					} else {
						$label = "<a name='@sections[$i]'>All @sections_display[$i]</a>";
						if ($one_big_page) { $button = "<a href=\#top>Top</a>"; }
						$space_after = 36;
					}
	
					$output .= "<tr><td colspan=6 bgcolor=#CCCCCC class=subheading><img src=../graphics/spacer.gif height=2 width=1><br>$label</td></tr>\r\r";
					
					# sort the all dates list by key and everything else by value
					if ((@sections[$i] =~ /dates/)&&($h == 1)) {
						@{@sections[$i]."_keys"} = sort (keys %{@sections[$i]});
						#@{@sections[$i]."_values"} = values %{@sections[$i]};
					} else {
						@{@sections[$i]."_keys"} = sort { ((${@sections[$i]}{$b} <=> ${@sections[$i]}{$a}) || ($a cmp $b)) } keys %{@sections[$i]};
						#@{@sections[$i]."_values"} = sort { int($b) <=> int($a) } values %{@sections[$i]};
					}
					
					# for the partial list, crop to the first $top_n values
					if ($h == 0) {
						splice(@{@sections[$i]."_keys"}, $top_n);
						#splice(@{@sections[$i]."_values"}, $top_n);
					}
						
					if ($show_performance) {
						$this_performance_time = time;
						$performance_times .= "&nbsp;&nbsp;sort and crop @sections[$i] data - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
						$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
						$last_performance_time = $this_performance_time;
					}

					# okay, now we can finally print the rows
					$this_limit = (@{@sections[$i]."_keys"} < $top_n) ? @{@sections[$i]."_keys"} : $top_n ;
					$keys_count = ($h > 0) ? @{@sections[$i]."_keys"} : $this_limit ;
					for ($j=0; $j < $keys_count; $j++) {
						
						$key = @{@sections[$i]."_keys"}[$j];
						#$value = @{@sections[$i]."_values"}[$j];
						$value = ${@sections[$i]}{$key};
						
						# prepare a link to allow inspecting an item in the list
						$inspector_key = @sections[$i];
						$inspector_value = $key;
						
						# separate the original value from the display value for the inspect script
						if (@sections[$i] == "visitors") {
							if (index($inspector_value, ") ") != -1) {
								$key = substr($inspector_value, 0, index($inspector_value, ") ") + 1);
								$inspector_value = substr($inspector_value, index($inspector_value, ") ") + 2);
							}
						}
						
						# untranslate the country code for the inspect script
						if (@sections[$i] =~ /countries/) {
							if (($inspector_value ne "")&&($inspector_value ne $blank_marker)) {
								$inspector_value = ".".@country_codes_reversed{$inspector_value};
							}
						}
						
						# do some basic URL-encoding on the inspect value
						$inspector_key =~ s/ /%20/g;
						$inspector_value =~ s/ /%20/g;
						$inspector_value =~ s/\=/%3D/g;
						$inspector_value =~ s/\&/%26/g;
						$inspector_value =~ s/\?/\\%3F/g;
						$inspector_value =~ s/\+/\\%2B/g;
						$inspector_value =~ s/\(/\\%28/g;
						$inspector_value =~ s/\)/\\%29/g;
						$inspector_value =~ s/\;/\\%3B/g;
						$inspector_value =~ s/\[/\\%5B/g;
						$inspector_value =~ s/\]/\\%5D/g;
	
						# limit the item display length to a predetermined number of characters
						$key_complete = $key;
						$key = (length($key) > $maximum_column_width) ? substr($key,0,$maximum_column_width)."..." : $key ;
						
						# make the page items into a link
						if (@sections[$i] =~ /pages/) {
							$key = ($key ne $blank_marker) ? "<a href=../assembler.cgi?page=$key_complete target='_new'>$key</a>" : $key ;
						}
				
						# make the IP address items into a link
						if (@sections[$i] =~ /IP_addresses/) {
							$key = ($key ne $blank_marker) ? "<a href=http://ws.arin.net/cgi-bin/whois.pl?queryinput=$key_complete target='_new'>$key</a>" : $key ;
						}
						
						# make the referrer items into a link
						if (@sections[$i] =~ /referrers/) {
							$key = ($key ne $blank_marker) ? "<a href=$key_complete target='_new'>$key</a>" : $key ;
						}
						
						# provide a link for seeing all the fields for a particular item
						$back_label = ($section) ? "All ".$section : "Top ".$top_n."%27s" ;
						$back_label =~ s/_/ /g;
						$back_label =~ s/ /%20/g;
						if ($show_inspector_rollover) {
							$inspector_link = ($key ne $blank_marker) ? "<a href='reporter.cgi?month=$month&inspector_key=$inspector_key&inspector_value=$inspector_value&back_label=$back_label'><img src=../graphics/inspector_0.gif onmouseover='document.images.inspector_$inspector_button_ID.src=\"../graphics/inspector_1.gif\"' onmouseout='document.images.inspector_$inspector_button_ID.src=\"../graphics/inspector_0.gif\"' width=12 height=12 border=0 name=inspector_$inspector_button_ID alt='Inspect'></a>" : "&nbsp;" ;
							$inspector_button_ID++;
						} else {
							$inspector_link = ($key ne $blank_marker) ? "<a href='reporter.cgi?month=$month&inspector_key=$inspector_key&inspector_value=$inspector_value&back_label=$back_label'><img src=../graphics/inspector_0.gif width=12 height=12 border=0 alt='Inspect'></a>" : "&nbsp;" ;					
						}
						
						$output .= "<tr valign=top><td class=body>\r";
						$output .= $inspector_link."&nbsp;";
						$output .= "</td>\n<td class=body>\r";
						$output .= $key;
						$output .= "</td>\n<td align=right><img src=../graphics/spacer.gif width=12 height=1></td>\n<td class=body>\r";
						$output .= $value;
						$output .= "</td>\n<td align=right><img src=../graphics/spacer.gif width=12 height=1></td>\n<td width=$graph_width>\r";
						$output .= "<font color=".random_color()."><img src=../graphics/spacer.gif height=4 width=".(($value / $page_views_greatest) * $graph_width)." border=1></font>\r";
						$output .= "</td></tr>\r\r";
					
					}
					
					# more/top buttons after listings
					$output .= "<tr><td colspan=6 class=body>$button</td></tr>\r\r";
					
					# space between rows
					if (!$section) { $output .= "<tr><td colspan=6><img src=../graphics/spacer.gif height=$space_after width=1></td></tr>\r\r"; }
				
					if ($show_performance) {
						$this_performance_time = time;
						$performance_times .= "&nbsp;&nbsp;print @sections[$i] data - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
						$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
						$last_performance_time = $this_performance_time;
					}

				}

			}
			
			# space between sections
			if ($h == 0) {
				if ($one_big_page) { $output .= "<tr><td colspan=6><img src=../graphics/spacer.gif height=36 width=1></td></tr>\r\r"; }
			}

		}
		
			
	# look for an error file and print those, too
		
		# skip if this is the menu and we're splitting the sections up
		if (($one_big_page)||($section)) {
			
			for ($h=0; $h<2; $h++) {
			
				$error_type = ($h == 0) ? "HTTP" : "assembler" ;
			
				# skip unless we're displaying this section, or no section is specified
				if ((!$section)||($section eq "errors_".$error_type)) {
	
					$error_list = "";

					$output .= "<tr><td colspan=6 bgcolor=#CCCCCC class=subheading><img src=../graphics/spacer.gif height=2 width=1><br><a name='errors_$error_type'>All $error_type errors</a></td></tr>\r\r";
			
					$text = get_file_simple($error_file);
					
					if (index($text, "Error: ") != 0) {
						@text = split("\r", $text);
						for ($line_index=0; $line_index<@text; $line_index++) {
							@next = split("\t", @text[$line_index], 500);
							if (@next == @columns_errors) { # make sure the line has all the elements we're expecting it to have
								
								# extract the values out of the log entry
								for ($i=0; $i < @columns_errors; $i++) {
									${"this_".@columns_errors[$i]} = (@next[$i] ne "") ? @next[$i] : $blank_marker ;
								}				
					
								# clean up some of the values
								$this_pages = (length($this_pages) > ($maximum_column_width + 24)) ? substr($this_pages, 0, ($maximum_column_width + 24))."..." : $this_pages ; # limit the page display length

								if (($error_type eq "assembler")&&(substr($this_message, 0, 3) !~ /\d\d\d/)&&(substr($this_message, 0, 3) !~ /Unk/)) {
									# show assembler errors first...
									$message =~ s/,/<br>/g;
									$error_list .= "<tr><td colspan=6 class=body>$this_dates<br>$this_pages<br>$this_message<hr></td></tr>\r";
								} elsif (($error_type eq "HTTP")&&((substr($this_message, 0, 3) =~ /\d\d\d/)||(substr($this_message, 0, 3) =~ /Unk/))) {
									# ...then HTTP errors
									$this_message =~ s/[^\d]//g; # strip it down to just the error code
									$this_message = (@HTTP_error_codes{$this_message}) ? "$this_message: @HTTP_error_codes{$this_message}" : "Unknown error" ;
									$error_list .= "<tr><td colspan=6 class=body>$dates<br>$pages<br>$this_message<hr></td></tr>\r";
								}
								
							}
						}
						$output .= $error_list;
					}
						
					if ($error_list) {
						# provide a button for clearing the error log
						$output .= "<tr><td colspan=6 class=body>\r";
						$output .= "<form name=clear method=post><input type=hidden name=clear value=1><input type=hidden name=month value=$month></form>\r";
						$output .= "<script language=JavaScript>\n function clearErrors() {\r";
						$output .= "if (confirm('Are you sure you want to clear the error log for this month?')) { document.clear.submit(); }\r }\r";
						$output .= "</script>\r";
						$output .= "<a href=JavaScript:clearErrors()>Clear error log</a>\r";
						$output .= "</td></tr>\r\r";
					} else {
						$output .= "<tr><td colspan=6 class=body>$blank_marker</td></tr>\r\r";
					}
					
					# more/top buttons after listings
					$output .= "<tr><td colspan=6 class=body>$button</td></tr>";
				
					# space between sections
					if ($h == 0) {
						if ($one_big_page) { $output .= "<tr><td colspan=6><img src=../graphics/spacer.gif height=36 width=1></td></tr>"; }
					}
	
				}
	
			}

		}		
		
		
	# finish off the table, and we're done
			
		$output .= "</table><br>\r\r";
		if (!$section) {
			$output .= "<p class=body>&lt;&ndash; <a href=reporter.cgi>Months</a></p>\r\r";
		} else {
			$output .= "<p class=body>&lt;&ndash; <a href=reporter.cgi?month=$month>Top ".$top_n."'s</a></p>\r\r";
		}
		

# otherwise, build a menu of all the available data files
	
	} else {
	
	
	# get a list of all the files in the data directory and sort it
	
		@data_files = ();
		
		opendir(dir, absolute_path($data_directory, "filesystem")) or
		$output .= "<b>Error:</b> Can't read the contents of the data directory $data_directory<p>";
		@data_files = grep {/^\d{4}-\d{2}\.txt$/} readdir(dir);
		$data_files_count = @data_files;
		closedir dir;
		
		@data_files = sort {$b cmp $a} (@data_files);
		
		
	# change each element in the data_files array to an array of information about the file

		$page_views_total = 0;
		$last_page_view = "NA";
		for ($i=0; $i < $data_files_count; $i++) {
			$name = @data_files[$i];
			@data_files[$i] = ();
			$data_files[$i]{"name"} = $name;

			$page_views = 0; # a simple count will do here
			@visits = (); %visits = (); # but here we have to remember each visit to avoid duplicates
			$time_cumulative = 0; # we'll count this in seconds and then average it later
			@visitors = (); %visitors = (); # here, again, we have to remember each visitor to avoid duplicates
			
			$text = get_file_simple($data_directory.$data_files[$i]{"name"});
			$data_files[$i]{"legacy_file"} = ($text =~ /(visitors not tracked)/); # this prevents us from displaying inaccurate visitor and visit data if we encounter a file that predates that functionality
			@text = split("\r", $text);
			@next = ();
			for ($line_index=0; $line_index<@text; $line_index++) {
				@next = split("\t", @text[$line_index], 500);
				if (@next == @columns) { # make sure the line has all the elements we're expecting it to have
			
					# extract the host value out of the log entry
					$this_hosts = ($track_visitors) ? @next[3] : @next[2] ;
					
					# check the host and skip this line if it matches our exclude hosts
					$include_this_host = 1;
					if (@exclude_from_hosts) {
						for ($k=0; $k < $exclude_from_hosts_count; $k++) {
							if ($this_hosts =~ /@exclude_from_hosts[$k]/i) { $include_this_host = 0; }
						}
					}

					# increment our various count variables
					if ($include_this_host) {
						$page_views++;
						if ($track_visitors) {
							# extract the date and visitor values out of the log entry
							$this_dates = $next[0];
							$this_visitors = $next[1];

							# don't track this at all if the visitor ID is missing (e.g. the client didn't accept the cookie)
							if (!$this_visitors) {
								$page_views--;
							} else {
							
								# counting visitors (unique visitor IDs this month) is easy
								# we're identifying visitors by the visitor ID
								$visitors{$this_visitors} = $this_visitors; # this automatically overwrites earlier entries with the same visitor, so we end up with a count of unique visitors
	
								# counting visits (page views by a visitor with no more than a $visit_expires minutes delay between them) is harder 
								if (!$visits{$this_visitors}) {
									# add a new visit entry with this time
									$visits{$this_visitors} = $this_dates; # we're identifying visits by the visitor ID and date
								} else {
									$last_dates = $visits{$this_visitors};
									$interval = (get_unix_time($this_dates) - get_unix_time($last_dates));
									if ($interval < ($visit_expires * 60)) {
										# if the last page view was less than $visit_expires minutes ago, update the visit entry with the time of this latest page view
										$visits{$this_visitors} = $this_dates;
										# also, add the time between these two page views to the cumulative visit duration
										$time_cumulative = $time_cumulative + $interval;
									} else {	
										# otherwise, tag the old visit entry with the end time for that visit and add a new visit entry with this new time
										$visits{"$this_visitors $last_dates"} = $last_dates;
										$visits{$this_visitors} = $this_dates;
										# note that we can't accurately say how long the user was on that last page, so we don't increment $time_cumulative here; this means our times will be lower than reality, but consistently lower; this seems to be the industry standard way to handle this problem
									}
								}
							
							}
						}
					}
					
				}
			}

			# do the math and save the results for this file into an array for display in a moment
			if ($page_views) {
				$data_files[$i]{"page_views"} = $page_views;
				if ($track_visitors) {
					@visits = keys(%visits); $visits = @visits;
					@visitors = keys(%visitors); $visitors = @visitors;
					$data_files[$i]{"visits"} = $visits;
					$data_files[$i]{"pages_per_visit"} = (int(($page_views / $visits * 100) + .5) / 100);
					$data_files[$i]{"time_per_visit"} = int(($time_cumulative / $visits) + .5);
					$data_files[$i]{"visitors"} = $visitors;
					$data_files[$i]{"visits_per_visitor"} = (int(($visits / $visitors * 100) + .5) / 100);
				}
			} else {
				$data_files[$i]{"page_views"} = 0;
				if ($track_visitors) {
					$data_files[$i]{"visits"} = 0;
					$data_files[$i]{"pages_per_visit"} = "NA";
					$data_files[$i]{"time_per_visit"} = "NA";
					$data_files[$i]{"visitors"} = 0;
					$data_files[$i]{"visits_per_visitor"} = "NA";
				}
			}
					
			# also keep track of the total page views for all files
			$page_views_total = $page_views_total + $page_views;
			if ($page_views > $page_views_greatest) { $page_views_greatest = $page_views; }
			
			# and get the last page view out of the first file, since we've sorted the files from newest to oldest
			if (($i == 0)&&(@next ne ())) {
				$last_page_view = @next[0];
			}
		}


	# print the file names and number of page views in a table
	
		$month_oldest = substr($data_files[$data_files_count - 1]{"name"}, 0, -4);

		$output .= "<p class=heading>Traffic since ".month_display($month_oldest)."</p>\r\r";
		
		$output .= "<p class=subheading>Last page view: $last_page_view</p>\r";
		$output .= "<p class=subheading>Total page views: $page_views_total</p><br>\r\r";
		
		$output .= "<table cellpadding=4 cellspacing=0 border=0>\r\r";
		
		$output .= "<tr bgcolor=#CCCCCC valign=bottom>\r";
		$output .= "<td class=subheading>Monthly&nbsp;reports</td>\r";
		$output .= "<td colspan=2 class=subheading>Page&nbsp;views</td>\r";
		if ($track_visitors) {
			$output .= "<td class=subheading>Visits</td>\r";
			$output .= "<td class=subheading>Pages/visit</td>\r";
			$output .= "<td class=subheading>Time/visit</td>\r";
			$output .= "<td class=subheading>Visitors</td>\r";
			$output .= "<td class=subheading>Visits/visitor</td>\r";
			$colspan = 8;
		} else {
			$colspan = 3;
		}
		$output .= "</tr>\r\r";

		for ($i=0; $i < $data_files_count; $i++) {
			$month = substr($data_files[$i]{"name"}, 0, -4);
			$output .= "<tr>\r";
			$output .= "<td class=body><a href=reporter.cgi?month=".$month.">".$month."</a></td>\r";
			$output .= "<td class=body align=right>".$data_files[$i]{"page_views"}."</td>\r";
			$output .= "<td width=$graph_width><font color=".random_color()."><img src=../graphics/spacer.gif height=4 width=".(($data_files[$i]{"page_views"} / $page_views_greatest) * $graph_width)." border=1></font></td>\r";
			if ($track_visitors) {
				if ($data_files[$i]{"legacy_file"}) {
					$output .= "<td class=body colspan=5>(visitors not tracked)</td>\r";
				} else {
					$time_per_visit_minutes = int($data_files[$i]{"time_per_visit"} / 60); # in Perl, int always rounds down, just like floor in the other languages
					$time_per_visit_seconds = $data_files[$i]{"time_per_visit"} % 60;
					$time_per_visit_seconds = sprintf("%02d", $time_per_visit_seconds);
					$time_per_visit_display = $time_per_visit_minutes.":".$time_per_visit_seconds;
					$output .= "<td class=body>".$data_files[$i]{"visits"}."</td>\r";
					$output .= "<td class=body>".$data_files[$i]{"pages_per_visit"}."</td>\r";
					$output .= "<td class=body>".$time_per_visit_display."</td>\r";
					$output .= "<td class=body>".$data_files[$i]{"visitors"}."</td>\r";
					$output .= "<td class=body>".$data_files[$i]{"visits_per_visitor"}."</td>\r";
				}
			}
			$output .= "</tr>\r\r";
		}
		
		$output .= "<tr><td colspan=$colspan>&nbsp;</td></tr>\r\r";

		$output .= "<tr><td colspan=$colspan bgcolor=#CCCCCC class=subheading>Other views</td></tr>\r\r";

		$output .= "<tr><td colspan=$colspan class=body>\r";	
		$output .= "<a href=\"JavaScript:window.open('realtime.swf?script_extension=cgi&graph_color=".$graph_color."&graph_image=".$graph_image."&update_interval=".$update_interval."','realtime','width=550,height=500,location=no,menubar=no,directories=no,toolbar=no,scrollbars=yes,resizable=yes,status=yes')\">Real-time display</a><br>\r\r";	
		$output .= "<a href=$data_directory>Raw logs</a><br>\r";	
		$output .= "</td></tr>\r\r";	
		
		$output .= "</table><br>\r\r";
	
	}
		
			
# print everything to the browser

	print $output;
	

# if requested, show a performance report

	if ($show_performance) {
		$this_performance_time = time;
		$performance_times .= "&nbsp;&nbsp;wrap up - ".($this_performance_time - $last_performance_time)." seconds <br>\r";
		$total_performance_time = $total_performance_time + ($this_performance_time - $last_performance_time);
		$last_performance_time = $this_performance_time;
	}

	if ($show_performance) {
		print "Language: cgi<br>\r";
		print "Host: $my_host<br>\r";
		print "Total page views: $page_views_total<br>\r";
		print "Performance times:<br>$performance_times\r";
		print "Total time: $total_performance_time seconds\r";
	}

# we're done!

	print "</div>\r\r";

	print "</body></html>\r\r";

	exit;
